/* gcc https.c -o https -lssl */

/* stdlib */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

/* socket */
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>

/* openssl */
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <signal.h>

SSL_CTX *ctx;

int create_socket(int port)
{
    int fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
        printf("socket() failed\n");
        return 0;
    }

    struct sockaddr_in hints = { 0 };
    hints.sin_family = AF_INET;
    hints.sin_port = htons(port);                 /* port number */
    hints.sin_addr.s_addr = htonl(INADDR_ANY);    /* internet address */

    if (bind(fd, (struct sockaddr *)&hints, sizeof(struct sockaddr_in)) < 0) {
        printf("bind() failed\n");
        close(fd);
        return 0;
    }
    if (listen(fd, 1) < 0) {
        printf("listen() failed\n");
        close(fd);
        return 0;
    }

    return fd;
}

void handle(SSL *ssl)
{
    char buffer[1024 * 8]; /* - 8kb buffer */

    /* receive */
    size_t read = SSL_read(ssl, buffer, sizeof(buffer) / sizeof(buffer[0]));

    /* isolate requested path */
    char *path = buffer;
    int i = 0;
    while (i++ < read && *path++ != ' ');
    while (i++ < read) {
        if (path[i] == ' ') {
            path[i] = '\0';
            break;
        }
    }
    if (path[i - 1] == '/') { /* redirect from root directory */
        const char headerBegin[] = "HTTP/1.1 301 Moved Permanently\nLocation: ";
        const char headerEnd[]   = "index.html\n";
        SSL_write(ssl, headerBegin, strlen(headerBegin));
        SSL_write(ssl, path, strlen(path));
        SSL_write(ssl, headerEnd, strlen(headerEnd));
        return;
    }
    /* make path relative */
     path--;
    *path = '.';

    /* respond */
    FILE *file = fopen(path, "r");
    if (file) {
        const char header[] = "HTTP/1.1 200 OK\n\n";
        if (SSL_write(ssl, header, strlen(header)) > 0) {
            while (   (read = fread(buffer, 1, sizeof(buffer) / sizeof(buffer[0]), file)) > 0
                    && SSL_write(ssl, buffer, read) == read);
        }
        fclose(file);
    } else {
        const char header[] = "HTTP/1.1 404 Not Found\n\n";
        SSL_write(ssl, header, strlen(header));
    }
}

static int volatile fd = 0;

static int volatile running = 1;

void sigint_handler(int signal) {
    running = 0;
    shutdown(fd, SHUT_RD);
}

int main()
{
    signal(SIGINT, sigint_handler);

    int err = 1;

    SSL_library_init();

    ctx = SSL_CTX_new(SSLv23_server_method());
    if (!ctx) {
        printf("SSL_CTX_new() failed\n");
        return err;
    }

    if (SSL_CTX_use_certificate_file(ctx, "cert.pem", SSL_FILETYPE_PEM) <= 0) {
        printf("SSL_CTX_use_certificate_file() failed\n");
        return err;
    }
    if (SSL_CTX_use_PrivateKey_file(ctx, "privkey.pem", SSL_FILETYPE_PEM) <= 0) {
        printf("SSL_CTX_use_PrivateKey_file() failed\n");
        return err;
    }

    int fd = create_socket(443);
    if (fd > 0) {
        while (running) {
            struct sockaddr_in client;
            socklen_t sockaddr_len = sizeof(struct sockaddr_in);
            /* accept socket */
            int connection = accept(fd, (struct sockaddr *)&client, &sockaddr_len);
            if (connection > 0) {
                /* accept ssl */
                SSL *ssl = SSL_new(ctx);
                if (ssl) {
                    SSL_set_fd(ssl, connection);
                    if (SSL_accept(ssl) > 0) {
                        handle(ssl);
                    }
                    SSL_free(ssl);
                }
                close(connection);
            }
        }

        err = 0;

        close(fd);
    }

    SSL_CTX_free(ctx);

    return err;
}
